/** \file SpvImage.h
 * Spv Image
 */


#ifndef SPV_2D_ARRAY_
#define SPV_2D_ARRAY_

struct Point2D {
	unsigned	mXPos;
	unsigned	mYPos;
};

template <class Type>
class	Spv2DArray
{
protected:
	Type				**mArrayPtr;
	unsigned			mWidth;
	unsigned			mHeight;

	unsigned			mGetPosX;
	unsigned			mGetPosY;
	unsigned			mGetStep;

	unsigned			mSetPosX;
	unsigned			mSetPosY;
	unsigned			mSetStep;

	void				DeleteMem(){
		if (mArrayPtr) {
			for(int i = 0; i < mHeight; i++)
				delete [] mArrayPtr[i];
			delete [] mArrayPtr;
		}
		mArrayPtr = NULL;
	}

	bool				CreateMem(){
		// Allocate the memory for the image
		if(!mArrayPtr) {
			mArrayPtr = new Type*[mHeight];
			for(int y = 0; y < mHeight; y++) {
				mArrayPtr[y] = new Type[mWidth];
			}

			return true;
		}
		return false;
	};

public:
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	///					Create Functions								///
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	Spv2DArray(){
		// Current Position
		// For the Get Functions
		mGetPosX	= 0;
		mGetPosY	= 0;
		mGetStep	= 1;

		// For the Set Functions
		mSetPosX	= 0;
		mSetPosY	= 0;
		mSetStep	= 1;

		// Set the Matrix 
		mArrayPtr	= NULL;
	}

	~Spv2DArray(){
		DeleteMem();
	}

	/// Create an empty array by a given Width and Height or a Dimension
	/// Can also be used for a collector of an array.
	void				CreateEmptyArray(unsigned width,unsigned height){
		// Delete the memory 
		if (mArrayPtr) DeleteMem();

		// Dimension of the image
		mWidth	= width;
		mHeight	= height;
		
		// Current Position
		mGetPosX = 0;
		mGetPosY = 0;
		mGetStep = 1;
		mSetPosX = 0;
		mSetPosY = 0;
		mSetStep = 1;

		mArrayPtr	= NULL;

		CreateMem();
	};
	void				CreateEmptyArray(Point2D dimension){
		CreateEmptyArray(Dimension.mYPos, Dimension.mXPos);
	};

	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	///					Iterator Functions								///
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	/// Initialize the iterator for the Get Next function
	bool				InitGetIterator(unsigned startPosX = 0, unsigned startPosY = 0, unsigned skipStep = 1){
		if(startPosX >= mWidth) {
			SPV_ERROR("Spv2DArray::InitGetIterator - "+ Itos(startPosX) + "(startPosX) >= " + Itos(mWidth) + "(mWidth)");
			return false;
		}
		if(startPosY >= mHeight) {
			SPV_ERROR("Spv2DArray::InitGetIterator - "+ Itos(startPosY) + "(startPosY) >= " + Itos(mHeight) + "(mHeight)");
			return false;
		}

		mGetPosX	= startPosX;
		mGetPosY	= startPosY;
		mGetStep	= skipStep;
		return true;
	};

	bool				InitGetIterator(Point2D startPos, unsigned skipStep = 1){
		return InitGetIterator(startPos.mXPos, startPos.mYPos, skipStep);
	};

	inline unsigned		GetGetSkipStep() {return mGetStep;};

	/// Gets the next value returns false if reached the end of the 2D Array
	bool				GetNext(Type &value){
		if(mGetPosY >= mHeight) {
			SPV_DEBUG("Spv2DArray::GetNext "+ Itos(mGetPosY) + "(mGetPosY) >= " + Itos(mHeight) + "(mHeight)", DBG_LVL_SPV);
			return false;
		}
		if(!GetVal(mGetPosX,mGetPosY,value)) 
			return false;

		mGetPosX += mGetStep;

		while(mGetPosX >= mWidth) {
			mGetPosX = mGetPosX - mWidth;
			if(mGetPosY < mHeight)
				mGetPosY += 1;
			else return true;
		}
		
		return true;
	};

	inline Point2D		GetGetPos() const {Point2D retPos = {mGetPosX,mGetPosY};	return retPos;};

	/// Initialize the iterator for the Set Next function
	bool				InitSetIterator(unsigned startPosX = 0, unsigned startPosY = 0, unsigned skipStep = 1){
		if(startPosX >= mWidth) {
			SPV_ERROR("Spv2DArray::InitSetIterator - "+ Itos(startPosX) + "(startPosX) >= " + Itos(mWidth) + "(mWidth)");
			return false;
		}
		if(startPosY >= mHeight) {
			SPV_ERROR("Spv2DArray::InitSetIterator - "+ Itos(startPosY) + "(startPosY) >= " + Itos(mHeight) + "(mHeight)");
			return false;
		}

		mSetPosX	= startPosX;
		mSetPosY	= startPosY;
		mSetStep	= skipStep;
		return true;
	};

	bool				InitSetIterator(Point2D startPos, unsigned skipStep = 1){
		return InitSetIterator(startPos.mXPos, startPos.mYPos, skipStep);
	};

	inline unsigned		GetSetSkipStep() {return mSetStep;};

	/// Sets the next value returns false if reached the end of the 2D Array
	bool				SetNext(Type value) {
		if(!SetVal(mSetPosX,mSetPosY,value))
			return false;

		mSetPosX += mSetStep;

		while(mSetPosX >= mWidth) {
			mSetPosX = mSetPosX - mWidth;
			if(mSetPosY < mHeight)
				mSetPosY += 1;
			else return true;
		}

		return true;
	};

	inline Point2D		GetSetPos() const {Point2D retPos = {mSetPosX,mSetPosY};	return retPos;};


	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	///					Get Functions									///
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	/// Get Dimension of the 2D Array
	Point2D				GetDimension()	const {Point2D retDim = {mWidth,mHeight};	return retDim;};

	/// Number of lines in the 2D Array
	inline unsigned		GetWidth() const	{return mWidth;};

	/// Number of columns in line
	inline unsigned		GetHeight() const	{return mHeight;};


	/// Returns a Section of the 2D Array
	virtual bool		GetSection(unsigned xLeft, unsigned yTop, unsigned xRight, unsigned yBottom, Spv2DArray &ret2DArray) const {
		unsigned xMin = (xRight > xLeft) ? xLeft : xRight;
		if(xMin >= mWidth) {
			SPV_ERROR("Spv2DArray::GetSection - " + Itos(xMin) + "(xMin) >= " + Itos(mWidth) + "(mWidth)");
			return false;  
		}
		unsigned xMax = (xRight > xLeft) ? xRight : xLeft;
		if(xMax > mWidth) xMax = mWidth;
		unsigned yMin = (yTop > yBottom) ? yBottom : yTop;
		if(yMin >= mHeight){
			SPV_ERROR("Spv2DArray::GetSection - " + Itos(yMin) + "(yMin) >= "+ Itos(mHeight) +"(mHeight)");
			return false;  
		}
		unsigned yMax = (yTop > yBottom) ? yTop : yBottom;
		if(yMax > mHeight) yMax = mHeight;

		ret2DArray.CreateEmptyArray(xMax-xMin, yMax-yMin);

		Type retVal;

		for(int y = yMin; y < yMax; y++){
			for(int x = xMin; x < xMax; x++) {
				if(!GetVal(xMin,yMin,retVal)) 
					return false;
				ret2DArray.SetVal(x-xMin,y-yMin,retVal);
			}
		}

		return true;
	};

	/// Returns a value in the 2D Array.  Returns false if error in index.
	virtual bool		GetVal(unsigned xPos, unsigned yPos, Type &value) const {
		if(!mArrayPtr) {
			SPV_ERROR("Spv2DArray::GetVal - 2D Array was not initiliazed correctly");
			return false;
		}

		if((xPos >= mWidth) || (yPos >= mHeight)) {
			SPV_ERROR("Spv2DArray::GetVal - Getting Value out of bound:\nx=" + Itos(xPos) +">="+ Itos(mWidth)+" or y=" + Itos(yPos) +">=" + Itos(mHeight));
			return false;
		}

		value = mArrayPtr[yPos][xPos];
		return true;
	};

	virtual bool		GetVal(Point2D xYPos, Type &value)  const {
		return GetVal(xYPos.mXPos, xYPos.mYPos, value);
	};

	/// Returns the number of elements in the 2D Array
	inline unsigned		GetElemCount() const {return mHeight * mWidth;};

	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	///					Set Functions									///
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	/// Set value to an element, report warning to the log file if in error Index.
	virtual bool		SetVal(unsigned xPos, unsigned yPos, Type value){
		if(!mArrayPtr) {
			SPV_ERROR("Spv2DArray::SetVal - 2D Array was not initiliazed correctly");
			return false;
		}

		if((xPos >= mWidth) && (yPos >= mHeight)) {
			SPV_ERROR("Spv2DArray::GetVal - Setting Value out of bound:\nx=" + Itos(xPos) +">="+ Itos(mWidth) + " or y=" + Itos(yPos) +">=" + Itos(mHeight));
			return false;
		}

		mArrayPtr[yPos][xPos] = value;
		return true;
	}

	virtual bool		SetVal(Point2D xYPos, Type value){
		return SetVal(xYPos.mXPos, xYPos.mYPos, value);
	};

	/// Set value to all the elements in the section
	bool				SetValueToSection(Point2D startXY, Point2D endXY, Type value){
		// Perform swap if needed (used mixed the values)
		if(yTop < yBottom) {
			unsigned tempVal = yBottom;
			yBottom = yTop;
			yTop = tempVal;
		}

		if(xLeft > xRight) {
			unsigned tempVal = xRight;
			xRight = xLeft;
			xLeft = tempVal;
		}

		for(;yBottom < yTop; yBottom++) {
			for(;xLeft < xRight; xLeft++) {
				if(!SetVal(xLeft,yBottom,value)) 
					return false;
			}
		}
		return true;
	};
	
	bool				SetValueToSection(unsigned xLeft, unsigned yTop, unsigned xRight, unsigned yBottom, Type value){
		return SetValueToSection(startXY.mXPos, startXY.mYPos, endXY.mXPos, endXY.mYPos,value);
	};
};

#endif

